import QtQuick 2.2
import QtQuick.Controls 1.1
import QtQuick.Layouts 1.1
import QtQuick.Dialogs 1.2
// FileDialog
import Qt.labs.folderlistmodel 2.1
import QtQml 2.2
import MuseScore 1.0
import FileIO 1.0
// This MuseScore Plugin is licensed under the GPL Version 2
//Copyright Joseph Eoff, April 2015

MuseScore {
    menuPath: "Plugins." + "Ultrastar Export"
    version: "0.0"
    description: "Export to Ultrastar format"

    onRun: {
        // check MuseScore version
        if (!(mscoreMajorVersion == 2 && mscoreMinorVersion >= 1)) {
            exportDialog.visible = false
            errorDialog.openErrorDialog(
                        qsTr("Minimum MuseScore Version 2.1.0 required for export"))
        }
        if (!(curScore)) {
            errorDialog.openErrorDialog(qsTr(
                                            "Select a score before exporting."))
            Qt.quit()
        } else {
            exportDialog.visible = true
            fillDefaultValues()
        }
    }

    function fillDefaultValues() {
        instrumentPlayer2.enabled = false
        voicePlayer2.enabled = false
        loadInstrumentList(staffList)
        instrumentPlayer1.model = staffList
        instrumentPlayer2.model = staffList
        loadVoiceList(instrumentPlayer1.currentText, player1Voices)
        loadVoiceList(instrumentPlayer2.currentText, player2Voices)
        exportDirectory.text = directorySelectDialog.fileUrl.toString().replace(
                    "file://", "")
    }

    function loadInstrumentList(instrumentList) {
        for (var i = 0; i < curScore.parts.length; i++) {
            var partname = curScore.parts[i].partName
            instrumentList.append({
                                      partname: partname
                                  })
        }
    }

    function loadVoiceList(instrumentName, voiceList) {
        for (var i = 0; i < curScore.parts.length; i++) {
            if (curScore.parts[i].partName === instrumentName) {
                voiceList.clear()
                for (var j = 0; j < curScore.parts[i].endTrack
                     - curScore.parts[i].startTrack; j++) {
                    voiceList.append({
                                         j: j
                                     })
                }
            }
        }
    }

    ListModel {
        id: staffList
    }

    ListModel {
        id: player1Voices
    }

    ListModel {
        id: player2Voices
    }

    MessageDialog {
        id: errorDialog
        visible: false
        title: qsTr("Error")
        text: "Error"
        onAccepted: {
            exportDialog.visible = false
            Qt.quit()
        }
        function openErrorDialog(message) {
            text = message
            open()
        }
    }

    MessageDialog {
        id: warningDialog
        visible: false
        title: qsTr("Warning")
        text: "Warning"
        onAccepted: {
            Qt.quit()
        }
        function openWarningDialog(message) {
            text = message
            open()
        }
    }

    FileIO {
        id: songTextWriter
        onError: console.log(msg + "  Filename = " + songTextWriter.source)
    }

    FileDialog {
        id: directorySelectDialog
        title: qsTr("Please choose a file")
        selectFolder: true
        visible: false
        onAccepted: {
            exportDirectory.text = directorySelectDialog.fileUrl.toString(
                        ).replace("file://", "")
        }
        Component.onCompleted: visible = false
    }

    Dialog {
        id: exportDialog
        visible: true
        title: qsTr("Ultrastar Export")
        width: formbackground.width
        height: formbackground.height
        contentItem: Rectangle {
            id: formbackground
            width: exporterColumn.width + 20
            height: exporterColumn.height + 20
            color: "lightgrey"
            ColumnLayout {
                id: exporterColumn
                GridLayout {
                    id: grid
                    columns: 2
                    anchors.left: parent.left
                    anchors.top: parent.top
                    anchors.margins: 10
                    Label {
                        text: qsTr("Player 1")
                    }
                    Label {
                        text: " "
                    }
                    Label {
                        text: qsTr("Instrument")
                    }
                    ComboBox {
                        id: instrumentPlayer1
                        onCurrentIndexChanged: {
                            loadVoiceList(currentText, player1Voices)
                            voicePlayer1.model = player1Voices
                        }
                    }
                    Label {
                        text: qsTr("Voice")
                    }
                    ComboBox {
                        id: voicePlayer1
                    }
                    Label {
                        text: qsTr("Duet")
                    }
                    CheckBox {
                        id: duet
                        onClicked: {
                            instrumentPlayer2.enabled = checked
                            voicePlayer2.enabled = checked
                        }
                    }
                    Label {
                        text: qsTr("Player 2")
                    }
                    Label {
                        text: " "
                    }
                    Label {
                        text: qsTr("Instrument")
                    }
                    ComboBox {
                        id: instrumentPlayer2
                        onCurrentIndexChanged: {
                            loadVoiceList(currentText, player2Voices)
                            voicePlayer2.model = player2Voices
                        }
                    }
                    Label {
                        text: qsTr("Voice")
                    }
                    ComboBox {
                        id: voicePlayer2
                    }
                    Label {
                        text: qsTr("Tempo")
                    }
                    TextInput {
                        id: tempoInput
                        text: "110"
                        inputMask: "000"
                    }
                    Button {
                        id: selectDirectory
                        text: qsTr("Select export directory")
                        onClicked: {
                            directorySelectDialog.open()
                        }
                    }
                    Label {
                        id: exportDirectory
                        text: ""
                    }
                    Button {
                        id: exportButton
                        text: qsTr("Export")
                        onClicked: {
                            exportUltrastar()
                            //Qt.quit()
                        } // onClicked
                    }
                    Button {
                        id: cancelButton
                        text: qsTr("Cancel")
                        onClicked: {
                            exportDialog.visible = false
                            Qt.quit()
                        } // onClicked
                    }
                }
                Label {
                    id: exportStatus
                    text: ""
                }
            }
        }
    }

    function exportUltrastar() {
        exportStatus.text = qsTr("Exporting .txt File.")
        if (!exportTxtFile()) {
            return false
        }
        exportStatus.text = qsTr("Exporting .mp3 File.")
        exportAudioFile()
        exportStatus.text = qsTr("")
    }

    function exportTxtFile() {
        var crlf = "\r\n"
        if (!(curScore.title)) {
            warningDialog.openWarningDialog(qsTr("Score must have a title."))
            return false
        }
        var filename = exportDirectory.text + "//" + filenameFromScore(
                    ) + ".txt"
        console.log(filename)

        var txtContent = ""
        txtContent += "#TITLE:" + curScore.title + crlf

        if (!(curScore.composer)) {
            warningDialog.openWarningDialog(qsTr("Score must have a composer."))
            return false
        }
        var tempo = 0
        tempo = parseInt(tempoInput.text)
        if (tempo < 1) {
            warningDialog.openWarningDialog(
                        qsTr("Tempo cannot be lower than 1 BPM"))
            return false
        }

        txtContent += "#ARTIST:" + curScore.composer + crlf

        txtContent += "#MP3:" + filenameFromScore() + ".mp3" + crlf
        txtContent += "#VIDEO:" + crlf
        txtContent += "#VIDEOGAP:" + crlf
        txtContent += "#START:0" + crlf
        txtContent += "#BPM:" + tempoInput.text + crlf
        var bpm = parseInt(tempoInput.text)
        //txtContent+="#BPM:"+curScore.tempo+crlf
        var cursor = getCursor(instrumentPlayer1.currentText,
                               voicePlayer1.currentText)
        var gap = 0 //getTimeOfFirstSyllable(cursor,bpm)
        txtContent += "#GAP:" + (gap * 1000) + crlf
        if (duet.checked) {
            txtContent += "P1" + crlf
        }
        cursor.rewind(0)
        txtContent += getSongText(cursor)

        if (duet.checked) {
            txtContent += "P2" + crlf
            cursor = getCursor(instrumentPlayer2.currentText,
                               voicePlayer2.currentText)
            txtContent += getSongText(cursor)
        }
        txtContent += "E" + crlf
        console.log(txtContent)
        songTextWriter.source = filename
        songTextWriter.write(txtContent)
        return true
    }

    function getTimeOfFirstSyllable(cursor, bpm) {
        while (cursor.next()) {
            if (cursor.element && cursor.element.type === Element.CHORD) {
                return ((cursor.tick / 490.96154) * 60 / bpm)
            }
        }
    }

    function getSongText(cursor) {
        var crlf = "\r\n"
        var syllable
        var songContent = ""
        var pitch_midi
        var timestamp_midi_ticks
        var duration_midi_ticks
        var lastMeasure
        var gotfirstsyllable = false
        var timeOfFirstSyllable = 0
        var lastItemWasMeasure = false

        while (cursor.next()) {
            if (cursor.measure !== lastMeasure) {
                timestamp_midi_ticks = calculateMidiTicksfromTicks(
                            cursor.tick, tempoInput.text)
                if (!lastItemWasMeasure && gotfirstsyllable) {
                    songContent += "-" + timestamp_midi_ticks + crlf
                    lastMeasure = cursor.measure
                }
                lastItemWasMeasure = true
            }
            if (cursor.element && cursor.element.type === Element.CHORD) {
                lastItemWasMeasure = false
                syllable = "__"
                if (cursor.element.lyrics.length > 0) {
                    syllable = cursor.element.lyrics[0].text
                    if (cursor.element.lyrics[0].syllabic === Lyrics.MIDDLE
                            || cursor.element.lyrics[0].syllabic === Lyrics.BEGIN) {
                        syllable += "-"
                    } else {
                        syllable += " "
                    }
                }
                pitch_midi = cursor.element.notes[0].pitch
                duration_midi_ticks = calculateMidiTicksfromTicks(
                            cursor.element.duration.ticks, tempoInput.text)
                timestamp_midi_ticks = calculateMidiTicksfromTicks(
                            cursor.tick, tempoInput.text)
                if (!gotfirstsyllable) {
                    gotfirstsyllable = true
                    timeOfFirstSyllable = timestamp_midi_ticks
                }
                songContent += ": " + timestamp_midi_ticks + " " + duration_midi_ticks
                        + " " + pitch_midi + " " + syllable + crlf
            }
        }
        return songContent
    }

    function calculateMidiTicksfromTicks(ticks, bpm) {
        return Math.round((ticks / 490.96154) * 60 / bpm * 15)
    }

    function getCursor(instrument, voice) {
        for (var i = 0; i < curScore.parts.length; i++) {
            if (curScore.parts[i].partName === instrument) {
                var track = curScore.parts[i].startTrack + parseInt(voice, 10)
                var cursor = curScore.newCursor(true)
                cursor.rewind(1)
                cursor.track = track
                cursor.rewind(0)
                return cursor
            }
        }
    }

    function getTempo() {//todo
    }

    function exportAudioFile() {
        var filename = exportDirectory.text + "//" + filenameFromScore()
        writeScore(curScore, filename, "mp3")
    }

    function filenameFromScore() {
        var name = curScore.title
        name = name.replace(/ /g, "_")
        return name
    }
}
